--[[
Name: uiQuery 1.0
Revision: $Rev: 28498 $
Website: http://www.wowace.com/
Description: ui wrapper
Dependencies: AceLibrary
]]

local MAJOR_VERSION = "uiQuery-1.0"
local MINOR_VERSION = "$Revision: 28498 $"

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local uiQuery, basefuncs = {}, {}
local events = {"OnAnimFinished", "OnChar", "OnClick", "OnCursorChanged", "OnDoubleClick", "OnDragStart", "OnDragStop", "OnEditFocusGained", "OnEditFocusLost", "OnEnter", "OnEnterPressed", "OnEscapePressed", "OnEvent", "OnHide", "OnHorizontalScroll", "OnHyperlinkClick", "OnHyperlinkEnter", "OnHyperlinkLeave", "OnInputLanguageChanged", "OnKeyDown", "OnKeyUp", "OnLeave", "OnMessageScrollChanged", "OnMouseDown", "OnMouseUp", "OnMouseWheel", "OnReceiveDrag", "OnScrollRangeChanged", "OnShow", "OnSizeChanged", "OnSpacePressed", "OnTabPressed", "OnTextChanged", "OnTextSet", "OnTooltipAddMoney", "OnTooltipCleared", "OnTooltipSetDefaultAnchor", "", "OnTooltipSetItem", "OnTooltipSetSpell", "OnTooltipSetUnit", "OnUpdate", "OnUpdateModel", "OnVerticalScroll", "OnValueChanged",}

local function __expSplit(exp)
    if exp:match(">") or exp:match(",") then
        local exp, op, exp2 = exp:match("(%w+) *(.) *(.+)")
        return exp, op, __expSplit(exp2)
    else
        return exp
    end
end

local function __expMerge(tbl, ...)
    for i = 1, select('#', ...) do
        tbl[#tbl + 1] = select(i, ...)    
    end
end

local function __expMatch(target, exp)
    local loose = exp:match("~(%w+)")
    if loose then 
        return target:IsObjectType(loose)
    else
        return target:GetObjectType() == exp
    end
end

local function __expSelect(exp, scope)
    if exp == '*' then return scope end
    frames = {}
    if scope[1].GetObjectType then
        for k, obj in pairs(scope) do
            if __expMatch(obj, exp) then table.insert(frames, obj) end
        end
    elseif scope[1][1] and scope[1][1].GetObjectType then
        for k, sscope in pairs(scope) do
            for k, obj in pairs(sscope) do
                if __expMatch(obj, exp) then table.insert(frames, obj) end
            end
        end
    else
        error("Invalid Scope", 2)
    end
    return frames
end

local function __expChildOf(exp, scope)
    local method, frames = "GetChildren"
    if exp == "FontString" or exp == "Texture" then method = "GetRegions" end
    if scope.GetChildren or scope.GetRegions then
        if exp == '*' then
            frames = {scope:GetChildren()} 
            __expMerge(frames, scope:GetRegions())
        else
            frames = __expSelect(exp, {scope[method](scope)})
        end
    elseif #scope == 0 then
        -- no bad monkey!
    else
        if exp == '*' then
            for k, sscope in pairs(scope) do
                if not frames then frames = {} end
                if sscope.GetChildren then __expMerge(frames, sscope:GetChildren()) end
                if sscope.GetRegions then __expMerge(frames, sscope:GetRegions()) end
            end
        else
            frames = {}
            for k, sscope in pairs(scope) do
                if sscope[method] then
                    for k, obj in pairs({sscope[method](sscope)}) do
                        if __expMatch(obj, exp) then table.insert(frames, obj) end
                    end
                end
            end
        end
    end
    return frames
end

local function __expStackWorker(stack, scope)
    if #stack == 0 then 
        return scope
    elseif #stack == 1 then
        error("Invalid Stack",2)
    else
        if stack[1] == ">" and stack[3] == "," then
            table.insert(stack, 2, "*")
            table.insert(stack, 3, ",")
        end
                
        local op, exp, frames = table.remove(stack, 1), table.remove(stack, 1)
      
        if op == ">" then
            return __expStackWorker(stack, __expChildOf(exp, scope))
        elseif op == "," then
            frames = {}
            while true do
                __expMerge(frames, unpack(__expSelect(exp, scope)))
                if #stack == 0 or stack[1] ~= "," then break end
                op, exp = table.remove(stack, 1), table.remove(stack, 1)
            end
            return __expStackWorker(stack, frames)
        end
    end
end

local function __exp(exp, scope, op)
    local stack, frames = {__expSplit(exp)}
    table.insert(stack, 1, op)
    return __expStackWorker(stack, scope)
end

for k, v in ipairs(events) do
    local script = v
    basefuncs[script] = function (self, func) 
        if func ~= nil then
            for k, frame in ipairs(self.__frames) do
                if frame:HasScript(script) then frame:SetScript(script, func or nil) end
            end
        elseif #self.__frames == 1 and self.__frames[1]:HasScript(script) then
            return self.__frames[1]:GetScript(script)
        end
        return self
    end
end

-- Helpers
function __sanctifyFrame(frame)
    if frame.isUiQuery then 
        if #frame.__frames > 1 then error("Ambigous argument passed in.  Query must reference only one object.", 3) end
        return frame.__frames[1] 
    else
        return frame
    end
end

function basefuncs:Center(offX, offY, relTo)
    for k, frame in ipairs(self.__frames) do
        if frame.SetPoint then frame:SetPoint("CENTER", relTo and __sanctifyFrame(relTo) or frame:GetParent(), "CENTER", offX or 0, offY or 0) end
    end
    return self
end

function basefuncs:Size(h, w, s)
    for k, frame in ipairs(self.__frames) do
        if not frame:IsObjectType("Region") then return end
        if not h and not w and not s and #self.__frames == 1 then
            return frame:GetHeight(), frame:GetWidth(), frame.GetScale and frame:GetScale() or nil
        else
            if h then frame:SetHeight(h) end
            if w then frame:SetWidth(w) end
            if s and frame.SetScale then frame:SetScale(s) end
        end
    end
    return self
end

function basefuncs:Filter(expression)
    if not expression then return self end
    return uiQuery:__createQuery(__exp(expression, self.__frames, ","), self)
end

function basefuncs:Children(exp)
    if not expression then expression = '*' end
    return uiQuery:__createQuery( __exp(expression, self.__frames, ">"), self)
end

function basefuncs:End()
    if not rawget(self, '__parentQuery') then return self else return self.__parentQuery end
end

local function __handler(self, ...)
    if select('#', ...) > 0 and select(1, ...) then
        local arg1 = ...
        return type(arg1[0]) == "userdata" and uiQuery:__createQuery({...}, self) or ...
    end
    return self 
end

local function __add(op1, op2)
    __expMerge(uiQuery:__createQuery(op1.__frames).__frames, unpack(op2.__frames))
    return op1
end

local function __index(self, key)
    local key = key
    
    self[key] = basefuncs[key] or function(self, arg1, arg2, arg3, ...)
        if arg1 and type(arg1) == "table" then arg1 = __sanctifyFrame(arg1) end
        if arg2 and type(arg2) == "table" then arg2 = __sanctifyFrame(arg2) end
        if arg3 and type(arg3) == "table" then arg3 = __sanctifyFrame(arg3) end
        if #self.__frames > 1 then
            local result, tmp
            for k, frame in ipairs(self.__frames) do
                if frame[key] then 
                    if not arg1 then
                        tmp = __handler(self, frame[key](frame))
                    elseif not arg2 then
                        tmp = __handler(self, frame[key](frame, arg1))
                    elseif not arg3 then
                        tmp = __handler(self, frame[key](frame, arg1, arg2))
                    else
                        tmp = __handler(self, frame[key](frame, arg1, arg2, arg3, ...))
                    end
                    if tmp and type(tmp) == "table" and tmp.isUiQuery then
                        if result then result = result + tmp else result = tmp end
                    end
                end
            end
            return result or self
        else
            if self.__frames[1][key] then
                if not arg1 then
                    return __handler(self, self.__frames[1][key](self.__frames[1]))
                elseif not arg2 then
                    return __handler(self, self.__frames[1][key](self.__frames[1], arg1))
                elseif not arg3 then
                    return __handler(self, self.__frames[1][key](self.__frames[1], arg1, arg2))
                else
                    return __handler(self, self.__frames[1][key](self.__frames[1], arg1, arg2, arg3, ...))
                end
            end
        end
    end

    return self[key]
end

function uiQuery:Create(type, name, parent, template)
    return self:__createQuery({CreateFrame(type, name, parent, template)})
end

function uiQuery:__call(...)
    if select('#', ...) == 0 then return self end
    local exp, scope = ...
    local frames
    
    if type(exp) == "table" then
        frames = {...}
    else
        if not scope or type(scope) ~= "table" or type(exp) ~= "string" or not scope:IsObjectType("Frame") then
            error("You must pass in a valid expression and scope to use this form of the constructor", 2)
        end
        
        frames = __exp(exp, scope, ">")
    end
    
    return self:__createQuery(frames)
end

function uiQuery:__index(key)
    if basefuncs[key] then return basefuncs[key] end
end

function uiQuery:__createQuery(frames, parent)
    if not frames or #frames == 0 then return self end
    if not self.registry[frames] then
        local query = { __frames = frames, __index = __index, __add = __add, isUiQuery = true}
        
        if parent then query.__parentQuery = parent end
        
        setmetatable(query, query)
        self.registry[frames] = query
    end
    
    return self.registry[frames]
end

local function activate(self, oldLib, oldDeactivate)
    uiQuery = self
    if oldLib then self.registry = oldLib.registry else self.registry = {} end    
    if oldDeactivate then oldDeactivate(oldLib) end
    setmetatable(uiQuery, uiQuery)
end

AceLibrary:Register(uiQuery, MAJOR_VERSION, MINOR_VERSION, activate)
uiQuery = AceLibrary(MAJOR_VERSION)